
# Panelization

When you want to panelize a board, you are expected to load the `kikit.panelize`
module and create an instance of the `Panel` class.

All units are in the internal KiCAD units (1 nm). You can use functions `fromMm(mm)` and
`toMm(kiUnits)` to convert to/from them (synopsis below). You are also encouraged to use
the functions and objects the native KiCAD Python API offers, e.g.: `wxPoint(args)`, `wxPointMM(mmx, mmy)`, `wxRect(args)`, `wxRectMM(x, y, wx, wy)`.



## Panel class

Basic interface for panel building. Instance of this class represents a
single panel. You can append boards, add substrate pieces, make cuts or add
holes to the panel. Once you finish, you have to save the panel to a file.
```
appendBoard(self, filename, destination, sourceArea=None, origin=Origin.Center, 
            rotationAngle=0, shrink=False, tolerance=0)
```
Appends a board to the panel.

The sourceArea (wxRect) of the board specified by filename is extracted
and placed at destination (wxPoint). The source area (wxRect) can be
auto detected if it is not provided. Only board items which fit entirely
into the source area are selected. You can also specify rotation. Both
translation and rotation origin are specified by origin. Origin
specifies which point of the sourceArea is used for translation and
rotation (origin it is placed to destination). It is possible to specify
coarse source area and automatically shrink it if shrink is True.
Tolerance enlarges (even shrinked) source area - useful for inclusion of
filled zones which can reach out of the board edges.

Returns bounding box (wxRect) of the extracted area.
```
save(self, filename)
```
Saves the panel to a file.
```
appendSubstrate(self, substrate, filletRadius=0)
```
Append a piece of substrate to the panel. Substrate can be either wxRect
or Shapely polygon. Newly appended corners can be rounded by specifying
non-zero filletRadius.
```
makeGrid(self, boardfile, rows, cols, destination, sourceArea=None, tolerance=0, 
         radius=0, verSpace=0, horSpace=0, verTabWidth=0, horTabWidth=0, 
         outerVerTabThickness=0, outerHorTabThickness=0)
```
Creates a grid of boards (row x col) as a panel at given destination
separated by V-CUTS. The source can be either extract automatically of
from given sourceArea. There can be a spacing between the individual
board (verSpacing, horSpacing) and the tab width can be adjusted
(verTabWidth, horTabWidth). Also the user can control whether append the
outer tabs (e.g. to connect it to a frame) by setting outerVerTabsWidth
and outerHorTabsWidth.

Returns a tuple - wxRect with the panel bounding box (excluding
outerTabs) and a list of cuts (list of lines) to make. You can use the
list to either create a V-CUTS via makeVCuts or mouse bites via
makeMouseBites.
```
makeTightGrid(self, boardfile, rows, cols, destination, verSpace, horSpace, 
              slotWidth, width, height, sourceArea=None, tolerance=0, radius=0, 
              verTabWidth=0, horTabWidth=0)
```
Creates a grid of boards just like `makeGrid`, however, it creates a
milled slot around perimeter of each board and 4 tabs.
```
makeFrame(self, innerArea, width, height, offset, radius=0)
```
Adds a frame around given `innerArea` (`wxRect`), which can be obtained,
e.g., by `makeGrid`, with given `width` and `height`. Space with width
`offset` is added around the `innerArea`.
```
makeVCuts(self, cuts)
```
Take a list of lines to cut and performs V-CUTS
```
makeMouseBites(self, cuts, diameter, spacing)
```
Take a list of cuts and perform mouse bites.
```
addNPTHole(self, position, diameter)
```
Add a drilled non-plated hole to the position (`wxPoint`) with given
diameter.

## Examples

### Simple grid

The following example creates 3 x 3 grid of boards in a frame separated by V-CUTS.

```
panel = Panel()
size, cuts = panel.makeGrid("test.kicad_pcb", 4, 3, wxPointMM(100, 40),
            tolerance=fromMm(5), verSpace=fromMm(5), horSpace=fromMm(5),
            outerHorTabThickness=fromMm(3), outerVerTabThickness=fromMm(3),
            verTabWidth=fromMm(15), horTabWidth=fromMm(8),
            radius=fromMm(1))
panel.makeVCuts(cuts)
# alternative: panel.makeMouseBites(cuts, diameter=fromMm(0.5), spacing=fromMm(1))
panel.makeFrame(size, fromMm(100), fromMm(100), fromMm(3), radius=fromMm(1))
panel.save("out.kicad_pcb")
```

